
INCLUDE "defines.asm"

INCLUDE "misc/rand.asm"
EXPORT randstate ; Defined in the above, exported here to avoid touching the file

SECTION "LCDMemsetSmallFromB", ROM0

; Writes a value to all bytes in an area of memory
; Works when the destination is in VRAM, even while the LCD is on
; @param hl Beginning of area to fill
; @param c Amount of bytes to write (0 causes 256 bytes to be written)
; @param a Value to write
; @return c 0
; @return hl Pointer to the byte after the last written one
; @return b Equal to a
; @return f Z set, C reset
LCDMemsetSmall::
	ld b, a
; Writes a value to all bytes in an area of memory
; Works when the destination is in VRAM, even while the LCD is on
; Protip: you may want to use `lb bc,` to set both B and C at the same time
; @param hl Beginning of area to fill
; @param c Amount of bytes to write (0 causes 256 bytes to be written)
; @param b Value to write
; @return c 0
; @return hl Pointer to the byte after the last written one
; @return b Equal to a
; @return f Z set, C reset
LCDMemsetSmallFromB::
	wait_vram
	ld a, b
	ld [hli], a
	dec c
	jr nz, LCDMemsetSmallFromB
	ret

; Writes a value to all bytes in a VRAM tilemap column
; Works when the destination is in VRAM, even while the LCD is on
; @param hl Address to the top of the VRAM tilemap column 
; @param c Amount of bytes to write (0 causes 256 bytes to be written)
; @param a Value to write
; @return c 0
; @return hl Pointer to the byte after the last written one
; @return b Equal to a
; @return f Z set, C reset
LCDMemsetSmallVertical::
	ld b, a
; Writes a value to all bytes in a VRAM tilemap column
; Works when the destination is in VRAM, even while the LCD is on
; Protip: you may want to use `lb bc,` to set both B and C at the same time
; @param hl Address to the top of the VRAM tilemap column 
; @param c Amount of bytes to write (0 causes 256 bytes to be written)
; @param b Value to write
; @return c 0
; @return hl Pointer to the byte after the last written one
; @return b unchanged
; @return f Z set, C reset
LCDMemsetSmallVerticalFromB::
	wait_vram
	ld  [hl],b       
	
	; add 32 to hl
	ld a, 32
	add l
	ld l, a
	adc h
	sub l
	ld h, a

	dec c

	jr  nz, LCDMemsetSmallVerticalFromB
	ret


SECTION "LCDMemset", ROM0

; Writes a value to all bytes in an area of memory
; Works when the destination is in VRAM, even while the LCD is on
; @param hl Beginning of area to fill
; @param bc Amount of bytes to write (0 causes 65536 bytes to be written)
; @param a Value to write
; @return bc 0
; @return hl Pointer to the byte after the last written one
; @return d Equal to parameter passed in a
; @return a 0
; @return f Z set, C reset
LCDMemset::
	ld d, a
; Writes a value to all bytes in an area of memory
; Works when the destination is in VRAM, even while the LCD is on
; @param hl Beginning of area to fill
; @param bc Amount of bytes to write (0 causes 65536 bytes to be written)
; @param d Value to write
; @return bc 0
; @return hl Pointer to the byte after the last written one
; @return a 0
; @return f Z set, C reset
LCDMemsetFromD::
	wait_vram
	ld a, d
	ld [hli], a
	dec bc
	ld a, b
	or c
	jr nz, LCDMemsetFromD
	ret

; Writes a value to all bytes in an squared section of a tilemap in VRAM
; Works when the destination is in VRAM, even while the LCD is on
; @param d Value to write
; @param hl Pointer to where to copy (address to top-left vertex of square)
; @param b Width of destination square
; @param c Height of destination square
LCDMemsetSquare::
	.prepare_c:
		push hl
		.loop_row
			wait_vram
			ld a, d
			ld [hli], a
			dec b
			jr nz, .loop_row

			pop hl					; sum 32 to hl
			; add 32 to hl
			ld a, 32
			add l
			ld l, a
			adc h
			sub l
			ld h, a

			dec c
			jr nz, .prepare_c
	ret

SECTION "LCDMemcpySmall", ROM0

; Copies a block of memory somewhere else in an horizontal fashion
; Works when the source or destination is in VRAM, even while the LCD is on
; @param de Pointer to beginning of block to copy
; @param hl Pointer to where to copy (bytes will be written from there onwards)
; @param c Amount of bytes to copy (0 causes 256 bytes to be copied)
; @return de Pointer to byte after last copied one
; @return hl Pointer to byte after last written one
; @return c 0
; @return a Last byte copied
; @return f Z set, C reset
LCDMemcpySmallHorizontal::
	wait_vram
	ld a, [de]
	ld [hli], a
	inc de
	dec c
	jr nz, LCDMemcpySmallHorizontal
	ret

SECTION "LCDMemcpy", ROM0

; Copies a block of memory somewhere else
; Works when the source or destination is in VRAM, even while the LCD is on
; @param de Pointer to beginning of block to copy
; @param hl Pointer to where to copy (bytes will be written from there onwards)
; @param bc Amount of bytes to copy (0 causes 65536 bytes to be copied)
; @return de Pointer to byte after last copied one
; @return hl Pointer to byte after last written one
; @return bc 0
; @return a 0
; @return f Z set, C reset
LCDMemcpy::
	wait_vram
	ld a, [de]
	ld [hli], a
	inc de
	dec bc
	ld a, b
	or c
	jr nz, LCDMemcpy
	ret

; Copies a block of memory somewhere else in  a vertical fashion
; Works when the source or destination is in VRAM, even while the LCD is on
; @param in HL - destination address of DATA (top of line in a VRAM tilemap)
; @param in B - width of source DATA
; @param in C - height of source DATA
; @param in DE - start of DATA
LCDMemcpyVertical::
	wait_vram
	ld  a,[de]      
	ld  [hl],a       

	;ld a, e
	;add b
	;ld e, a

	;ld a, d
	;adc 0
	;ld d, a

	; add 32 to de
	ld a, b
	add e
	ld e, a
	adc d
	sub e
	ld d, a
	
	; add 32 to hl
	ld a, 32
	add l
	ld l, a
	adc h
	sub l
	ld h, a

	dec c

	jr  nz, LCDMemcpyVertical
	ret

; Copies a square section from Rom to a tilemap in VRAM
; Works when the source or destination is in VRAM, even while the LCD is on
; @param de Pointer to beginning of block to copy
; @param hl Pointer to where to copy (bytes will be written from there onwards)
; @param b width of square
; @param c height of square
; @param a width of source data
; @return de Pointer to byte after last copied one
; @return hl Pointer to byte after last written one
; @return c 0
; @return f Z set, C reset

LCDMemcpySquare::
	sub b
	.prepare_c:
		push af
		push bc
		ld c, a

		push hl
		.loop_row
			wait_vram
			ld a, [de]
			ld [hli], a
			inc de
			dec b
			jr nz, .loop_row
			
			ld a, e					; sum width to de
			add c
			ld e, a 
			ld a, d
			adc 0
			ld d, a 

			pop hl					
			; add 32 to hl
			ld a, 32
			add l
			ld l, a
			adc h
			sub l
			ld h, a


			pop bc
			pop af

			dec c
			jr nz, .prepare_c
	ret

SECTION "Memcpy", ROM0

; Copies a block of memory somewhere else
; @param de Pointer to beginning of block to copy
; @param hl Pointer to where to copy (bytes will be written from there onwards)
; @param bc Amount of bytes to copy (0 causes 65536 bytes to be copied)
; @return de Pointer to byte after last copied one
; @return hl Pointer to byte after last written one
; @return bc 0
; @return a 0
; @return f Z set, C reset
Memcpy::
	ld a, [de]
	ld [hli], a
	inc de
	dec bc
	ld a, b
	or c
	jr nz, Memcpy
	ret

SECTION "SpriteRoutines", ROM0

; Copies sprite struct from ROM (GO####) to RAM (wGO####) and sprite OAM data (placed just after GO####) from ROM to Shadow OAM (wGO####OAM)
; @param de Pointer to GameObject data in ROM (GO####)
; @param hl Pointer to GameObject assigned address in RAM (wGO####)
; @param bc Pointer to GameObject assigned address in OAM (wGO####OAM)
; @return de Pointer to last position of sprite OAM data in ROM
; @return hl Pointer to the last attribute of last sprite in OAM
; @return c 0
; @return f Z set, C reset
LoadSprite::
	push bc

    ld c, sizeof_GameObject
    rst MemcpySmall

	pop hl
	ld a, [de]
	ld c, a
	inc de
	rst MemcpySmall

	ret

; Toggles a OAM flag in a collection of sprites
; @param hl Pointer to the OAM flags of first sprite
; @param c OAM flag to flip
; @param b Number of sprites
; @return hl Pointer to the OAM flags of last sprite
; @return b 0
TOGGLE_OAM_FLAG::
	inc l
	inc l
	inc l
	.loop 
		ld a, [hl]
		xor c
		ld [hl], a

		dec  b

		ret   z

		inc l
		inc l
		inc l
		inc l
		jr .loop
		ret

; Swaps the 4 top sprites with the 4 bottom sprites (need to be placed adjacent, only works with 2x4 sprites)
; @param hl Pointer to byte 4 of last sprite
; @param b Number of sprites
; @return hl Pointer to byte 4 of first sprite
; @param b 0
ChangeTopBottom::
	dec hl
	.loop  
		ld a, [hl]
		xor %0001000
		ld [hld], a

		dec hl
		dec hl
		dec hl

		dec  b
		jr   nz,.loop

		ret

; Sets the X position for a collection of sprites
; @param hl Pointer to the x attribute of first sprite
; @param c Where to place game object
; @param d Amount of sprites horizontal
; @param e Amount of sprites vertical
; @return hl Pointer to the last sprite moved
; @return b initial d (Amount of sprites horizontal)
; @return f Z set, C reset
SetGameObjectPositionX::
	ld a, c
	ld b, d
	.loop_width:
		ld [hl], a
		dec d
		jr z, .loop_height

		inc l
		inc l
		inc l
		inc l
		add 8
		jr .loop_width
	.loop_height:
		dec e
		ret z
		ld d, b
		ld a, c
		inc l
		inc l
		inc l
		inc l
		jr .loop_width

	ret

; Sets the Y position for a collection of sprites
; @param hl Pointer to the y attribute of first sprite
; @param c Where to place game object
; @param d Amount of sprites horizontal
; @param e Amount of sprites vertical
; @return hl Pointer to the last sprite moved
; @return b initial d (Amount of sprites horizontal)
; @return f Z set, C reset
SetGameObjectPositionY::
	ld a, c
	ld b, d
	.loop_width:
		ld [hl], a
		dec d
		jr z, .loop_height

		inc l
		inc l
		inc l
		inc l
		jr .loop_width
	.loop_height:
		add 16
		dec e
		ret z
		ld d, b
		inc l
		inc l
		inc l
		inc l
		jr .loop_width

	ret

; Scrolls a collection of sprites to either x or y direction
; @param hl Pointer to the first sprite to be moved (wGO#OAM_sprite_1_1_y to move up and down and wGO#OAM_sprite_1_1_x to move left and right)
; @param b Amount of sprites to move
; @param c Amount of pixels by which to move each sprite	
; @return hl Pointer to the last sprite moved
; @return b 0
; @return f Z set, C reset
ScrollGameObject::
	ld a, [hl]
	add a, c
	ld [hl], a

	dec  b
	ret   z

	inc l	; works because we're working with OAM, where the LOW byte is always C1
	inc l 
	inc l 
	inc l 

	jr ScrollGameObject
	ret

SECTION "OtherRoutines", ROM0

; Call the routine a pointer is pointing to
; @param a Value of pointer (offset)
; @param hl Pointer to the routine array
; @return f Z set, C reset
; @destroy a
CallRoutinePointerArray::
	add a  ; multiply A by 2 because each pointer is 2 bytes long
	add l
	ld l, a
	adc h
	sub l
	ld h, a  ; HL points to the desired function pointer
	ld a, [hl+]
	ld h, [hl]
	ld l, a  ; HL contains the function pointer
	
	rst CallHL
	ret